# Type system

Typon reuses the notational conventions from [PEP 484](https://www.python.org/dev/peps/pep-0484/), and the types defined in the `typing` module.

The type system as it is now is probably not sound, but this is something that will evolve in the future.

## Algebraic Data Types

Typon supports ADTs in the form of tuples (product types) and unions (sum types):

```py
from typing import Optional

a = (1, "a", True)  # tuple[int, str, bool]

def f(x: Optional[int] = None):  # alias for Union[int, NoneType]
    pass
    
def g(y: str | int):  # alias for Union[str, int]
    pass
```

## Inference

Currently, a derivative of Hindley-Milner is used to infer types. A unification-based system has some disadvantages though,
notably non-locality of errors and the difficulty of modeling inheritance, and as such this will probably be changed to
something based on [bidirectional typing](https://arxiv.org/pdf/1908.05839.pdf).

However, even with the current system, the type inference is quite powerful, and Python programs can be compiled often with no changes at all,
with an immediate performance bump.